import os
from pathlib import Path
import numpy as np


def check_rm(path):
    """
    Check and delete file

    :param path: path string of target file
    """
    if os.path.isfile(path):
        print(f"removing {path}")
        os.remove(path)


def file_exists(path):
    """
    Check if file exists. If not, raise FileNotFoundError.

    :param path: path string of target file
    :returns: True if file exists
    """
    if os.path.isfile(path):
        return True
    else:
        raise FileNotFoundError(f"{path} not exist")


def is_empty(path):
    """
    Check if file is empty. If the file does not exist, then create one.

    :param path: path string of target file
    :returns: boolean operator
    """
    if os.path.isfile(path):
        return os.stat(path).st_size == 0
    else:
        Path(path).touch()

def efficient_log(data, replace_zero=10):
    """A shortcut to efficiently calculate log10 of data.

    :param data: data to calculate log10
    :param replace_zero: value to replace zero in data

    :returns: log10 of data
    """
    return np.where(data == 0, replace_zero, np.log10(data))



import xml.etree.ElementTree as ET
from importlib.resources import files
import itertools

import numpy as np
import pandas as pd
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import xarray as xr

import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, to_rgb as hex_to_rgb, LinearSegmentedColormap, rgb2hex
import matplotlib.patheffects as pe
from matplotlib.patheffects import Stroke, Normal
import cartopy.mpl.geoaxes
from cgeniepy.grid import GridOperation
from .utils import efficient_log
import warnings


class GriddedDataVis:

    """A class to visualise the GriddedData object"""

    transform_crs = ccrs.PlateCarree()  # do not change

    def __init__(self, gd):
        """
        initialise the GriddedDataVis object

        :param gd: GriddedData object
        """
        self.data = gd.data
        self.attrs = gd.attrs
        
        ## aesthetic parameters
        self.aes_dict = {
            ## x,ylabel
            "general_kwargs": {"cmap": "viridis", "font": "Helvetica", "fontsize": 10},
            "facecolor_kwargs": {"c": "silver"},
            "borderline_kwargs": {"c": "black", "linewidth": 0.5},
            "outline_kwargs": {"colors": "black", "linewidth": 0.5},
            "gridline_kwargs": {"color": "black", "linewidth": 0.5, "linestyle": "dashed", "draw_labels": False},
            "pcolormesh_kwargs": {"shading": "auto", "cmap": plt.get_cmap("viridis")},
            "contour_kwargs": {
                "linewidths": 0.6,
                "colors": "black",
                "linestyles": "solid",
                "zorder": 10,
                "levels": 15,
            },
            "contour_label_kwargs": {
                "colors": ["black"],
                "fontsize": 8,
                "inline": False,
            },
            "contourf_kwargs": {"levels": 15},
            "colorbar_label_kwargs": {
                "label": f"{self.attrs['long_name']}\n({self.attrs['units']})",
                "size": 10,
                "labelpad": 10,
            },
            "colorbar_kwargs": {"fraction": 0.046, "pad": 0.03},
        }

    def plot(self, *args, **kwargs):
        """
        visualise the data based on the dimension of the array

        *args and **kwargs are passed to plotting functions
        in practical, turn on/off the plotting elements
        e.g., plot(x, y, pcolormesh=False, contour=True)
        each plotting function will seek for the corresponding kwargs
        in self.aes_dict
        """
        plt.rcParams['font.family'] = 'sans-serif'

        ## try to update pcolor_kwargs
        if 'vmin' in kwargs:
                self.aes_dict['pcolormesh_kwargs']['vmin'] = kwargs['vmin']
        if 'vmax' in kwargs:
            self.aes_dict['pcolormesh_kwargs']['vmax'] = kwargs['vmax']
        if 'cmap' in kwargs:
            self.aes_dict['pcolormesh_kwargs']['cmap'] = kwargs['cmap']

                   
        

        if self.data.ndim == 1:
            return self._plot_1d(*args, **kwargs)
        elif self.data.ndim == 2:
            return self._plot_2d(*args, **kwargs)
        elif self.data.ndim == 3:
            return self._plot_3d(*args, **kwargs)
        else:
            raise ValueError(f"{self.data.ndim} dimensions not supported")

    def _plot_1d(self, *args, **kwargs):
        """
        plot 1D data, e.g., zonal_average, time series
        dim: time/lon/lat
        """
        if "ax" not in kwargs:
            fig, local_ax = self._init_fig(figsize=(5, 3.5))
        else:
            local_ax = kwargs.pop("ax")

        ## get the only dimension as x
        dim = self.data[self.data.dims[0]]
        
        p = local_ax.plot(dim, self.data, *args, **kwargs)
            
        
        local_ax.grid(True, linestyle='--', alpha=0.5)
        local_ax.minorticks_on()
        local_ax.tick_params(axis='both', which='both', direction='in', top=True, right=True)        
        local_ax.set_xlabel(dim.name)
        local_ax.set_ylabel(f"{self.attrs['long_name']} ({self.attrs['units']})")

        return p

    def _plot_2d(self, *arg, **kwargs):
        "plot 2D data, e.g., map, trasect"
        ## if lon, lat then plot map
        ## if zt or depth then plot transec   

        has_lat, has_lon, has_depth, has_time = GridOperation().check_dimension(self.data.dims)
        if has_lat and has_lon:
            return self._plot_map(*arg, **kwargs)
        elif has_depth:
            return self._plot_transect(*arg, **kwargs)
        else:
            raise ValueError("2D plotting not supported")

    def _plot_3d(self, *args, **kwargs):
        "plot 3D data = plot mutiple 2D plots"
        has_lat, has_lon, has_depth, has_time = GridOperation().check_dimension(self.data.dims)
        
        if has_time or has_depth:
            target_order = GridOperation().dim_order(self.data.dims)[0]
            target_name = self.data.dims[target_order]
            target_arr = self.data[target_name]

            # Determine the optimal number of columns and rows for subplots
            num_plots = len(target_arr)
            ncols = min(num_plots, 3)
            nrows = (num_plots + ncols - 1) // ncols  # Calculate the required number of rows

            fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=(3 * ncols, 2.2 * nrows),
                                    sharex=True, sharey=True, squeeze=False)

            # Flatten the axes array for easier iteration
            axs = axs.flatten()

            # Plot the data for each time step                
            for i, ax in enumerate(axs):
                if i < num_plots:
                    im = self.data.isel({target_name: i}).plot(ax=ax, add_colorbar=False)
                    title = f"{target_name} = {target_arr[i].values:.2f} {self.data[target_name].attrs['units']}"
                    ax.set_title(title, fontsize=10)
                    ax.set_xlabel("")
                    ax.set_ylabel("")
                else:
                    ax.set_visible(False)

            # Add a common colorbar to the figure
            fig.colorbar(im, ax=axs.tolist(), orientation='horizontal', label=f"{self.attrs['long_name']} ({self.attrs['units']})",
                         fraction=0.046, pad=0.05)            
        else:            
            raise ValueError("Not support 3D plot iterating over other dimension than time and depth")
    

    def _plot_map(
        self,
        pcolormesh=True,
        contour=False,
        colorbar=False,
        contourf=False,
        outline=False,
        facecolor=True,
        borderline=True,
        gridline=False,
        cfeature=None,
        zebra_frame=False,
        *args,
        **kwargs,
    ):

        dim_order = GridOperation().dim_order(self.data.dims)
        lat_order = dim_order[0] ## in the case of 2D, lat is the first dimension
        lon_order = dim_order[1] ## in the case of 2D, lon is the second dimension

        x_name = self.data.dims[lon_order]  ## lon
        y_name = self.data.dims[lat_order]  ## lat

        x_arr = self.data[x_name].values
        y_arr = self.data[y_name].values

        x_min = x_arr.min()
        x_max = x_arr.max()
        x_res = x_arr[1] - x_arr[0]
        x_edge = np.linspace(x_min-x_res/2, x_max+x_res/2, x_arr.size + 1)


        y_edge = np.rad2deg(np.arcsin(np.linspace(-1, 1, y_arr.size + 1)))


        if "ax" not in kwargs:
            fig, local_ax = self._init_fig(subplot_kw={"projection": ccrs.EckertIV()})
        else:
            local_ax = kwargs.pop("ax")

        if facecolor:
            self._set_facecolor(local_ax, **self.aes_dict["facecolor_kwargs"])

        if borderline:
            self._set_borderline(
                local_ax, geo=True, **self.aes_dict["borderline_kwargs"]
            )

        if outline:
            ## outline uses edge coordinates
            ## need to be transformed to PlateCarree
            self._add_outline(
                local_ax,
                x=x_edge,
                y=y_edge,
                transform=self.transform_crs,
                **self.aes_dict["outline_kwargs"],
            )

        if gridline:
            self._add_gridline(
                local_ax,
                transform=self.transform_crs,
                **self.aes_dict["gridline_kwargs"],
            )


        if pcolormesh:
            ## pcolormesh uses edge coordinates
            ## need to be transformed to PlateCarree
            p_pcolormesh = self._add_pcolormesh(
                local_ax,
                x=x_edge,
                y=y_edge,
                transform=self.transform_crs,
                *args,
                **self.aes_dict["pcolormesh_kwargs"],
            )
            if colorbar:
                cbar = self._add_colorbar(p_pcolormesh, orientation="horizontal", **self.aes_dict["colorbar_kwargs"])
                self._add_colorbar_label(cbar, **self.aes_dict["colorbar_label_kwargs"])

        if contour:
            ## contour uses center coordinates
            ## need to be transformed to PlateCarree
            p_contour = self._add_contour(
                local_ax,
                x_arr,
                y_arr,
                transform=self.transform_crs,
                **self.aes_dict["contour_kwargs"],
            )
            self._add_contour_label(
                local_ax, p_contour, **self.aes_dict["contour_label_kwargs"]
            )
            ## contour will not be used to plot colorbar because it's set to black

        if contourf:
            p_contourf = self._add_contourf(
                local_ax,
                x_arr,
                y_arr,
                transform=self.transform_crs,
                **self.aes_dict["contourf_kwargs"],
            )

        if cfeature:
            self._add_cfeature(local_ax, cfeature)
        
        if zebra_frame:
            add_zebra_frame(local_ax)

        if pcolormesh:
            return p_pcolormesh
        elif contour:
            return p_contour
        elif contourf:
            return p_contourf
        else:
            return local_ax

    def _plot_transect(
        self,
        pcolormesh=True,
        contour=False,
        colorbar=True,
        contourf=False,
        outline=False,
        facecolor=True,
        borderline=True,
        *args,
        **kwargs,
    ):
        """
        Examples
        ---------------

        import matplotlib.pyplot as plt

        model = EcoModel("path_to_model")

        fig, axs=plt.subplots(nrows=1, ncols=3, figsize=(15, 3))

        basins = ['Atlantic', 'Pacific', 'Indian']

        for i in range(3):
            model.get_var('ocn_PO4').isel(time=-1).mask_basin(base='worjh2',basin=basins[i], subbasin='').mean(dim='lon').plot(ax=axs[i])
            axs[i].title.set_text(basins[i])


        available kwargs:
        outline_kwargs = {'outline_color': 'red', 'outline_width': .5}
        """
        ## sort the dimension
        dim_order = GridOperation().dim_order(self.data.dims)        

        zt_order = dim_order[0] ## in the case of 2D, zt is the first dimension
        lat_order = dim_order[1] ## in the case of 2D, lat is the second dimension
        
        
        x_name = self.data.dims[lat_order]  ## lat
        y_name = self.data.dims[zt_order]  ## zt

        x_arr = self.data[x_name].values
        y_arr = self.data[y_name].values

        x_edge = np.rad2deg(np.arcsin(np.linspace(-1, 1, x_arr.size + 1)))
        ## get y edge coordinates (starting from 0)
        ## this assumes depth is the mid point of two edges
        y_edge = np.zeros(len(y_arr)+1)
        y_edge[1] = y_arr[0] * 2

        for i in range(1, len(y_arr)):    
            half_length = y_arr[i] - y_edge[i]
            next_edge = y_arr[i] + half_length
            y_edge[i+1] = next_edge        

        
        if "ax" not in kwargs:
            fig, local_ax = self._init_fig(figsize=(5, 3))
        else:
            local_ax = kwargs.pop("ax")

        if facecolor:
            self._set_facecolor(local_ax, **self.aes_dict["facecolor_kwargs"])

        if borderline:
            self._set_borderline(
                local_ax, geo=False, **self.aes_dict["borderline_kwargs"]
            )

        if outline:
            ## outline uses edge coordinates
            self._add_outline(
                local_ax, x=x_edge, y=y_edge, **self.aes_dict["outline_kwargs"]
            )

        if pcolormesh:
            ## pcolormesh uses edge coordinates
            p_pcolormesh = self._add_pcolormesh(
                local_ax, x=x_edge, y=y_edge, *args, **self.aes_dict["pcolormesh_kwargs"]
            )
            if colorbar:
                cbar = self._add_colorbar(p_pcolormesh, orientation="vertical")
                self._add_colorbar_label(cbar, **self.aes_dict["colorbar_label_kwargs"])

        if contour:
            ## contour uses center coordinates
            ## need to be transformed to PlateCarree
            p_contour = self._add_contour(
                local_ax, x_arr, y_arr, **self.aes_dict["contour_kwargs"]
            )
            self._add_contour_label(
                local_ax, p_contour, **self.aes_dict["contour_label_kwargs"]
            )
            ## contour will not be used to plot colorbar because it's set to black


        if contourf:
            p_contourf = self._add_contourf(local_ax, x_arr, y_arr, **self.aes_dict["contourf_kwargs"])

        ## reverse y axis
        local_ax.set_ylim(local_ax.get_ylim()[::-1])
        local_ax.set_ylabel(
            y_name,
            fontsize=self.aes_dict["general_kwargs"]["fontsize"],
            font=self.aes_dict["general_kwargs"]["font"],
        )
        local_ax.set_xlabel(
            x_name,
            fontsize=self.aes_dict["general_kwargs"]["fontsize"],
            font=self.aes_dict["general_kwargs"]["font"],
        )

        ## x/y tick label
        local_ax.tick_params(
            axis="both",
            which="major",
            direction="out",
            labelsize=self.aes_dict["general_kwargs"]["fontsize"],
            labelfontfamily=self.aes_dict["general_kwargs"]["font"],
        )
        local_ax.minorticks_on()

        if pcolormesh:
            return p_pcolormesh
        elif contour:
            return p_contour
        elif contourf:
            return p_contourf
        else:
            return local_ax


    ## ------- Below is implementations -------------------------

    def _init_fig(self,dpi=120, *args, **kwargs):
        return plt.subplots(dpi=dpi, *args, **kwargs)

    def _add_pcolormesh(self, ax, x, y, *args, **kwargs):
        return ax.pcolormesh(x, y, self.data, *args, **kwargs)

    def _add_contour(self, ax, x, y, *args, **kwargs):
        return ax.contour(x, y, self.data, *args, **kwargs)

    def _add_contourf(self, ax, x, y, *args, **kwargs):
        return ax.contourf(x, y, self.data, *args, **kwargs)

    def _add_contour_label(self, ax, cs, *args, **kwargs):
        ax.clabel(cs, cs.levels[::2], *args, **kwargs)

    def _add_gridline(self, ax, *args, **kwargs):
        gl = ax.gridlines(*args, **kwargs)        

        if ax.projection == ccrs.PlateCarree():
            gl.xlines = False  # removing gridlines
            gl.ylines = False
            add_border_ticks(ax, 0.015)


    def _set_borderline(self, ax, geo=True, **kwargs):
        if geo:
            ax.spines["geo"].set_edgecolor(kwargs["c"])
            ax.spines["geo"].set_linewidth(kwargs["linewidth"])
        else:
            for direction in ["top", "bottom", "left", "right"]:
                ax.spines[direction].set_edgecolor(kwargs["c"])
                ax.spines[direction].set_linewidth(kwargs["linewidth"])
                # vertical layer order
                ax.spines[direction].set_zorder(0)

    def _set_facecolor(self, ax, **kwargs):
        ax.patch.set_color(**kwargs)

    def _add_outline(self, ax, x, y, **kwargs):
        """
        draw outlines for the NA grids
        """
        mask_array = np.where(~np.isnan(self.data), 1, 0)

        Ny_edge = len(y)
        Nx_edge = len(x)
        # dimension of array
        Ny_array = Ny_edge - 1
        Nx_array = Nx_edge - 1
        # index of array
        Ny_index = Ny_array - 1
        Nx_index = Nx_array - 1

        # i stands for yitude index, j stands for xgitude index
        # e.g., mask_array[0, ] is Antarctic ice cap
        # i,j are for mask array and will be converted to x/y edge lines
        for i in range(Ny_array):
            for j in range(Nx_array):
                # compare with the right grid, and plot vertical line if different
                if j < Nx_index and mask_array[i, j] != mask_array[i, j + 1]:
                    ax.vlines(x[j + 1], y[i], y[i + 1], **kwargs)

                # connect the circular xgitude axis
                if j == Nx_index and mask_array[i, j] != mask_array[i, 0]:
                    ax.vlines(x[j + 1], y[i], y[i + 1], **kwargs)

                # compare with the above grid, and plot horizontal line if different
                if i < Ny_index and mask_array[i, j] != mask_array[i + 1, j]:
                    ax.hlines(y[i + 1], x[j], x[j + 1], **kwargs)

    def _add_colorbar(self, mappable_object, *args, **kwargs):
        cbar = plt.colorbar(mappable_object, *args, **kwargs)
        return cbar

    def _add_colorbar_label(self, cbar, *args, **kwargs):
        ## set colorbar label
        cbar.set_label(*args, **kwargs)
        cbar.ax.tick_params(
            axis="both",
            which="major",
            labelsize=self.aes_dict["general_kwargs"]["fontsize"],
            labelfontfamily=self.aes_dict["general_kwargs"]["font"],
        )
        cbar.outline.set_edgecolor("black")
        cbar.outline.set_linewidth(0.5)

    def _add_cfeature(self, ax, feature_name):
        """add natural earth mask"""

        feature = cfeature.NaturalEarthFeature('physical', feature_name, '110m',
                                        edgecolor='black', facecolor='silver')
        ## use ax's transform
        ax.add_feature(feature, zorder=10)



class ScatterDataVis:
    "Visualisation object based on ScatterData object"

    def __init__(self, sd):
        self.data = sd.data.reset_index(inplace=False)
        self.index = sd.index

        gp = GridOperation()
        gp.set_coordinates(obj=self, index=self.index)

    def _init_fig(self, *args, **kwargs):
        return plt.subplots(*args, **kwargs)

    def plot(self, var, *args, **kwargs):
        """
        visualise the data based on the dimension of the data
        """
        plt.rcParams['font.family'] = 'sans-serif'

        if isinstance(self.index, str):
            self.index = [self.index]

        match len(self.index):
            case 1:
                return self._plot_1d(var=var, *args, **kwargs)
            case 2:
                if hasattr(self, 'lat') and hasattr(self, 'lon'):
                    ## drop kind if it exsit in kwargs
                    del kwargs['kind']
                    warnings.warn("kind is not used for map plotting")
                    return self._plot_map(var=var, *args, **kwargs)
                elif hasattr(self, 'depth') and hasattr(self, 'lat'):
                    del kwargs['kind']
                    warnings.warn("kind is not used for transect plotting")
                    return self._plot_transect(var=var,*args, **kwargs)
                else:
                    raise ValueError("self.dim must be lat/lon or depth not real column name")
            case _:
                raise ValueError("Not supported 3D and higher dimensions")

    def _plot_1d(self, var, kind='scatter', *args, **kwargs):
        """
        plot 1D data, e.g., zonal_average, time series
        dim: time/lon/lat
        """
        if "ax" not in kwargs:
            fig, local_ax = self._init_fig()
        else:
            local_ax = kwargs.pop("ax")

        ## get the only dimension as x
        x = self.data[self.index[0]]
        y = self.data[var]

        if kind == 'line':
            p = local_ax.plot(x, y, *args, **kwargs)
        elif kind == 'scatter':
            p = local_ax.scatter(x, y, *args, **kwargs)
        else:
            raise ValueError("Invalid plot type. Choose 'line' or 'scatter'.")
        
        local_ax.grid(True, linestyle='--', alpha=0.5)
        local_ax.minorticks_on()
        local_ax.tick_params(axis='both', which='both', direction='in', top=True, right=True)

        return p


    def _plot_map(
        self,
        var,
        ax=None,
        log=False,
        land_mask=True,
        colorbar=True,
            gridline=True,
            zebra_frame=False,
        *args,
        **kwargs,
    ):
        """plot map based on dataframe with latitude/longitude
        using cartopy as engine

        :param x: coordinate attribute, default "Longitude"
        :param y: coordinate attribute, default "Latitude"
        :param interpolate: whether interpolate scatter data

        :returns: a map
        """

        if not ax:
            fig, ax = self._init_fig(subplot_kw={"projection": ccrs.PlateCarree()})

        if land_mask:
            ax.set_global()
            ## plot land and coastline, zorder is the drawing order, smaller -> backer layer
            # ax.stock_img()
            ax.add_feature(cfeature.LAND.with_scale("110m"), zorder=2, facecolor="lightgrey")
            ax.add_feature(cfeature.COASTLINE.with_scale("110m"), linewidth=1.4)
            ax.add_feature(cfeature.LAKES.with_scale('110m'), zorder=3, facecolor='black')

        if log:
            self.data[var] = efficient_log(self.data[var])

        if self.data[var].dtype != float:
           self.data[var] = self.data[var].astype(float)

        p = ax.scatter(
            self.data[self.lon],
            self.data[self.lat],
            c=self.data[var],
            transform=ccrs.PlateCarree(),
            *args,
            **kwargs,
        )

        if colorbar:
            cbar = plt.colorbar(p, ax=ax, orientation="horizontal",
                                label=f"{var}", pad=0.1)
            cbar.ax.tick_params(axis="both", which="major", direction="out", length=5)
            cbar.ax.minorticks_on()

        if zebra_frame:
            add_zebra_frame(ax)

        if gridline:
            gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True, linewidth=.5,
                              color='black', alpha=0.5,
                              linestyle='--', zorder=1)
            gl.top_labels = False
            gl.right_labels = False
            if isinstance(ax.projection, ccrs.PlateCarree):
                add_border_ticks(ax, 0.01)

        return p    

    def _plot_transect(
        self,
        var,
        ax=None,
        bathy_lon=None,
        *args,
        **kwargs,
    ):
        """plot transect based on dataframe with depth/latitude
        bathy_lon: longitude range for the bathymetry plot
        """
        if not ax:
            fig, ax = self._init_fig()

        if self.data[var].dtype != float:
            self.data[var] = self.data[var].astype(float)
            
        p = ax.scatter(
            x=self.data[self.lat],
            y=self.data[self.depth],
            c=self.data[var],
            *args,
            **kwargs,
        )        

        if bathy_lon:
            data_path = str(files('data').joinpath('GEBCO2002_bathy.nc'))
            bathy = xr.open_dataset(data_path)

            ## get the lat range
            min_lat = self.data[self.lat].min()
            max_lat = self.data[self.lat].max()
            min_depth = self.data[self.depth].min()

            up = bathy.z.sel(lon=slice(*bathy_lon)).mean(dim="lon")
            # up = up.sel(lat=slice(min_lat, max_lat))
            bottom = np.ones(len(up)) * -5500

            ax.fill_between(up.lat, up, bottom, color="black")

        return p


class CommunityPalette:

    """A class to handle community-driven colormaps"""

    def __init__(self, name=None, *args, **kwargs):
        self.name = name
        
        if self.name:
            self.colormap = self.get_palette(name, *args, **kwargs)
        else:
            self.colormap=None

    def get_palette(self, cmap_name, N=256, reverse=False, alpha=None):
        """
        community-driven colormaps with multiple sources

        :param cmap_name: colormap name, can be found in avail_palette()
        :type cmap_name: str

        :returns: colormap

        XML data: https://sciviscolor.org/colormaps/
        txt data: from original packages
        """

        ## if _r exsit in the end of the string
        if cmap_name[-2:] == "_r":
            cmap_name = cmap_name[:-2]
            reverse = True

        if cmap_name not in self.avail_palettes():
            raise ValueError(
                f"{cmap_name} not found, accepted values are {self.avail_palettes()}"
            )

        data_dir = files("data").joinpath("colormaps")

        file_path = None
        file_ext = None
        colors = []
        c = None

        file_path = [f for f in data_dir.glob("**/*") if f.stem == cmap_name]        
        file_path = file_path[0]
        file_ext = file_path.suffix


        match file_ext:
            case ".txt":
                colors = self._parse_txt_data(file_path)

                if alpha is not None:
                    rgba_colors = [(*hex_to_rgb(color), alpha) for color in colors]
                    c = ListedColormap(rgba_colors, name=cmap_name)
                else:
                    c = ListedColormap(colors, name=cmap_name)
                    
            case ".xml":
                colors = self._parse_xml_data(file_path)
                if alpha is not None:
                    rgba_colors = [(color[0], color[1], color[2], alpha) for color in colors]                    
                    c = ListedColormap(rgba_colors, name=cmap_name)
                else:
                    c = ListedColormap(colors, name=cmap_name)
            case ".spk":
                data = self._parse_spk_data(file_path)
                pos = data[:,0]
                ## normalise pos to 0-1
                pos_min = np.min(pos)
                pos_max = np.max(pos)
                scaled_pos = [(val - pos_min) / (pos_max - pos_min) for val in pos]

                rgb = data[:,1:]/100

                ## if alpha is not None, add alpha to the rgb
                if alpha is not None:
                    rgba = np.zeros((len(rgb), 4))
                    rgba[:,0:3] = rgb
                    rgba[:,3] = alpha
                    c = self.create_colormap(scaled_pos, rgba)
                else:
                    c = self.create_colormap(scaled_pos, rgb)
            case _:
                raise ValueError("File extension not supported")
            
        interval = np.linspace(0, 1, N)
        c = ListedColormap(c(interval), name=cmap_name)
        
        if reverse and c is not None:
            return c.reversed()

        if c is None:
            raise ValueError("Colormap could not be created")

        return c

    def avail_palettes(self, show_ferret_data=True):
        """return a list of colormap names"""
        data_dir = files("data").joinpath("colormaps")

        if show_ferret_data:
            return [
                f.stem
                for f in data_dir.glob("*")
                if f.suffix in [".txt", ".xml", ".spk"]
            ]
        else:
            return [
                f.stem
                for f in data_dir.glob("*")
                if f.suffix in [".txt", ".xml"]
            ]

    def _parse_spk_data(self, filename):
        data = []
        with open(filename, 'r') as file:
            for line in file:
                # Skip empty lines and comment lines starting with "!"
                if "RGB_Mapping" in line or not line.strip() or line.strip().startswith('!'):
                    continue
                # Split the line by whitespace and convert to floats
                if "!" in line:
                    line = line.split("!")[0]
                values = list(map(float, line.split()))
                data.append(values)
        return np.array(data)

    def _parse_txt_data(self, filename):
        colors = pd.read_csv(filename, header=None).values.tolist()
        colors = [colors[i][0] for i in range(len(colors))]
        return colors

    def _parse_xml_data(self, filename):
        tree = ET.parse(filename)
        root = tree.getroot()
        colors = []
        for point in root.findall(".//Point"):
            r = float(point.get("r"))
            g = float(point.get("g"))
            b = float(point.get("b"))
            colors.append((r, g, b))
        return colors
        

    def to_rgb(self):
        return self.colormap.colors

    def to_hex(self, unique=True):
        df = pd.DataFrame(self.colormap.colors)
        ## get unique rows
        if unique:
            df = df.drop_duplicates()
            
        x_unique = df.to_numpy()

        hex_color = [rgb2hex(i) for i in x_unique]
        return hex_color

    
    def create_colormap(self, positions, colors):
        """
        Create a colormap with specified positions and colors.

        Args:
            positions (list): List of floats indicating the positions of colors in the colormap.
            colors (list): List of RGB tuples representing the colors.

        Returns:
            LinearSegmentedColormap: The created colormap.
        """
        segments = {'red': [], 'green': [], 'blue': []}
        for pos, color in zip(positions, colors):
            for i, key in enumerate(['red', 'green', 'blue']):
                segments[key].append((pos, color[i], color[i]))

        cmap = LinearSegmentedColormap('custom_colormap', segments)
        return cmap


    def __repr__(self):
        x = np.linspace(0, 1, 256).reshape(1,-1)
        plt.imshow(x, cmap=self.colormap, aspect=20)
        plt.axis('off')
        plt.show()
        ## title

        return f"Colormap: {self.name}"


def add_zebra_frame(ax, lw=1.2):
    """
    copied from https://github.com/SciTools/cartopy/issues/1830
    """
    ax.spines["geo"].set_visible(False)
    left, right, bot, top = ax.get_extent()

    # Alternate black and white line segments
    bws = itertools.cycle(["k", "white"])

    xticks = sorted([left, *ax.get_xticks(), right])
    xticks = np.unique(np.array(xticks))
    yticks = sorted([bot, *ax.get_yticks(), top])
    yticks = np.unique(np.array(yticks))
    for ticks, which in zip([xticks, yticks], ["lon", "lat"]):
        for idx, (start, end) in enumerate(zip(ticks, ticks[1:])):
            bw = next(bws)
            if which == "lon":
                xs = [[start, end], [start, end]]
                ys = [[bot, bot], [top, top]]
            else:
                xs = [[left, left], [right, right]]
                ys = [[start, end], [start, end]]

            # For first and lastlines, used the "projecting" effect
            capstyle = "butt" if idx not in (0, len(ticks) - 2) else "projecting"
            for (xx, yy) in zip(xs, ys):
                ax.plot(
                    xx,
                    yy,
                    color=bw,
                    linewidth=lw,
                    clip_on=False,
                    transform=ccrs.PlateCarree(),
                    solid_capstyle=capstyle,
                    # Add a black border to accentuate white segments
                    path_effects=[
                        pe.Stroke(linewidth=lw + 1, foreground="black"),
                        pe.Normal(),
                    ],
                )

def add_border_ticks(ax, tick_len_scale=0.015):
    """
    copied from https://github.com/SciTools/cartopy/issues/2003
    """
    ## add ticks
    xticks = np.linspace(-180, 180, 7)
    yticks = np.linspace(-90, 90, 7)
    ax.set_xticks(xticks, crs=ccrs.PlateCarree())
    ax.set_yticks(yticks, crs=ccrs.PlateCarree())
    ax.set_yticklabels("")
    ax.set_xticklabels("")
    bbox_pixels = ax.get_window_extent().get_points()
    width_pixels = bbox_pixels[1, 0] - bbox_pixels[0, 0]

    # Set the tick length as a fraction of the width
    tick_length = tick_len_scale * width_pixels            
    ax.tick_params(axis='x', length=tick_length)
    ax.tick_params(axis='y', length=tick_length)

    



from cgeniepy.skill import ArrComparison
import numpy as np

def create_testdata():
    x = np.linspace(0,100,100)
    y = x    

    ## calculate skill score
    return ArrComparison(x, y)    

def test_mscore():    
    ac = create_testdata()
    assert ac.mscore()==1.0

def test_pearson_r():
    ac = create_testdata()
    diff = ac.pearson_r().item() - 1.0
    assert diff < 1E-8

def test_cos_sim():    
    ac = create_testdata()
    assert ac.cos_similarity()==1.0

def test_rmse():
    ac = create_testdata()
    assert ac.rmse()==0.0



import cgeniepy.chem as chem

def test_molcularweight():
    assert chem.Chemistry().molecular_weight('CO2') == 44.0095


def test_formatunit():
    input = 'mmol C m-2 yr-1'
    assert chem.Chemistry().format_unit(input) == 'mmol C m$^{-2}$ yr$^{-1}$'


from matplotlib.testing.decorators import image_comparison
from cgeniepy.table import ScatterData
from importlib.resources import files
import cgeniepy
import matplotlib.pyplot as plt
import cartopy.crs as ccrs

def create_sample_data():
    model = cgeniepy.sample_model()
    return model.get_var('ocn_sur_temp').isel(time=-1)

@image_comparison(baseline_images=['test_map'], remove_text=True,
                  extensions=['png'], style='mpl20')
def test_map():    
    fig, ax = plt.subplots(subplot_kw={'projection': ccrs.Mollweide()})
    data = create_sample_data()
    data.plot(ax=ax)
    
    return fig

@image_comparison(baseline_images=['test_line'], remove_text=True,
                  extensions=['png'], style='mpl20')
def test_line():
    data = create_sample_data()
    fig, ax = plt.subplots()
    data.mean(dim='lon').plot(ax=ax)
    return fig

@image_comparison(baseline_images=['test_scatterdatavis'], remove_text=True,
                  extensions=['png'], style='mpl20')
def test_scatterdatavis():
    file_path = str(files("data").joinpath("EDC_CO2.tab"))
    data= ScatterData(file_path, sep='\t')
    data.set_index('Age [ka BP]')
    fig, ax = plt.subplots()
    data.plot(var='CO2 [ppmv]', ax=ax)
    return fig


import numpy as np
from cgeniepy.skill import DFComparison
import geopandas as gpd
from shapely.geometry import Point
import pandas as pd

from io import StringIO
import re
from typing import Union

from cgeniepy.grid import Interpolator, GridOperation

from cgeniepy.plot import ScatterDataVis
from cgeniepy.grid import GridOperation
import cgeniepy.array as ca
from importlib.resources import files


class ScatterData:
    """ScatterData is a class to store non-gridded data with columns of coordinates."""

    def __init__(self, data: Union[pd.DataFrame, int, str], mutable: bool = False, **kwargs):
        """
        Initialize a ScatterData object.

        Parameters:
        data: The path to the file, the data, or a PanDataSet ID.
        mutable: Whether the data is mutable.
        **kwargs: Additional keyword arguments for pandas read functions.
        """
        self.mutable = mutable
        self.data = self._process_data(data, **kwargs)

        # if the index is already set in the data, then set the coordinates
        if not isinstance(self.data.index, pd.core.indexes.range.RangeIndex):
            self.index= list(self.data.index.names)
            GridOperation().set_coordinates(obj=self, index=self.index)

    def _process_data(self, data: Union[pd.DataFrame, int, str], **kwargs) -> pd.DataFrame:
        if isinstance(data, pd.DataFrame):
            return data
        if isinstance(data, int):
            try:
                from pangaeapy.pandataset import PanDataSet
                return PanDataSet(data).data
            except ImportError:
                print("Unable to import PanDataSet from pangaeapy. Please make sure the package is installed.")                            
        if isinstance(data, str):
            return self._process_string_data(data, **kwargs)
        raise ValueError("Unsupported data type. Expected DataFrame, int, or str.")

    def _process_string_data(self, data: str, **kwargs) -> pd.DataFrame:
        if data.endswith(".tab"):
            return pd.read_csv(StringIO(self._parse_tab_file(data)), **kwargs)
        if data.endswith(".xlsx"):
            return pd.read_excel(data, **kwargs)
        if "PANGAEA" in data:
            try:
                from pangaeapy.pandataset import PanDataSet
            except ImportError:
                print("Unable to import PanDataSet from pangaeapy. Please make sure the package is installed.")   

            doi = self._extract_doi(data)
            return PanDataSet(doi).data
        return pd.read_csv(data, **kwargs)

    @staticmethod
    def _extract_doi(url: str) -> str:
        doi_pattern = r"10\.\d{4,9}/[-._;()/:A-Z0-9]+"
        doi_match = re.search(doi_pattern, url, re.IGNORECASE)
        if doi_match:
            return doi_match.group(0)
        raise ValueError("No valid DOI found in the URL.")

    def __repr__(self):
        prefix = "ScatterData\n"
        columns = f"Columns: {self.data.columns}\n"
        rows = f"Rows: {len(self.data)}\n"
        if hasattr(self, "index"):
            index = f"Index: {self.index}\n"
        else:
            index = ""

        return prefix + columns + index + rows

    def _parse_tab_file(self, filename, begin_cmt = '/*', end_cmt = '*/'):
        """
        Read a tab-delimited file and return a pandas that is optimised for pangea-format data.

        :param filename: The name of the file to read.
        :param begin_cmt: The string that marks the beginning of a comment.
        :param end_cmt: The string that marks the end of a comment.

        :return: A pandas dataframe.
        """
        lines = []
        in_comment = False

        with open(filename) as f:
            for line in f:
                if line.strip().startswith(begin_cmt):
                    in_comment = True
                    continue

                if line.strip().startswith(end_cmt):
                    in_comment = False
                    continue

                if not in_comment:
                    lines.append(line.rstrip('\n'))

        data = '\n'.join(lines)
        return data    

    def __getitem__(self, item):
        return self.data[item]
    
    def _check_cols(self, cols):
        """
        Check if the columns are present in the dataframe.

        :param cols: A list of columns to check.

        :raises: ValueError if a column is not found.
        """
        for col in cols:
            if col not in self.data.columns:
                raise ValueError(f"{col} not found in the dataframe")


    def set_index(self, index):
        """Tell the object which columns are the coordinates.        
        """
        self.data.set_index(index, inplace=True)
        self.index= index
        GridOperation().set_coordinates(obj=self, index=self.index)


    def reset_index(self):
        if self.mutable:
            self.data = self.data.reset_index()
            return self
        else:
            return self.data.reset_index()
        
    def detect_basin(self):
        """use point-in-polygon strategy to detect modern ocean basin according to lon/lat column

        Example
        ----------
        >>> data = ScatterData(data)
        >>> data.set_index(['lon', 'lat'])
        >>> data.detect_basin()
        """
        
        def sub_detect_basin(lon, lat):
            p = Point(lon, lat)
            ocean_name = oceans[oceans.contains(p)].Oceans.values
            if ocean_name.size > 0:
                return ocean_name[0]
            else:
                return ""
            
        file_path = str(files("data").joinpath("oceans/oceans.shp"))
        oceans = gpd.read_file(file_path)
        
        if len(self.index) != 2:
            raise ValueError("The index must have two columns: lon and lat")
        
        assert self.lon in self.index, f"{self.lon} not found in the index"
        assert self.lat in self.index, f"{self.lat} not found in the index"

        result = self.data.reset_index().apply(lambda row: sub_detect_basin(row[self.lon], row[self.lat]), axis=1)
        

        ## add to self.data
        if self.mutable:
            self.data['basin'] = result
            return self
        else:        
            tmp_data = self.data.reset_index()
            tmp_data['basin'] = result
            tmp_index = self.index
            y = ScatterData(tmp_data)
            y.set_index(tmp_index)
            return y


    def to_xarray(self):
        "convert to xarray dataset"
        ## set the coordinate using the data from self.coordinates         
        return self.data.to_xarray()

    def to_GriddedData(self, var):
        "convert to gridded data"
        output = ca.GriddedData(self.to_xarray()[var])
        return output

    def interpolate(self, var):
        ## a tuple of coordinate arrays
        coords =  tuple([self.data.reset_index(inplace=False)[dim].values for dim in self.index])
        ## array values
        values = self.data[var].values
        dims = self.index

        ## if 1D
        if len(dims) == 1:            
            output= Interpolator(dims, coords, values, 200, '1d').to_dataframe()
            output.rename(columns={"interpolated_values": var}, inplace=True)
            output.set_index(dims, inplace=True)
        else:
            output= Interpolator(dims, coords, values, 200, 'ir-linear')

        if self.mutable:
            self.data = output
            return self
        else:
            return output


    def to_geniebin(
        self,
        var,
        agg_method="mean"
    ):
        """
        Regrid a dataframe within certain format to cGENIE grids.
        This method does not consider the land-sea mask. So if convert from a higher resolution data to GENIE grids,
        the land-sea mask could be different (Not a problem in comparison though).

        :param var: The variable to regrid.
        :param agg_method: The aggregation method to use when regridding.
        :return: An indexed data frame
        """

        go= GridOperation()

        src_df = self.data.reset_index(inplace=False)

        # subset
        select_cols = self.index + [var]
        src_df = src_df[select_cols]

        # drop NAN
        src_df = src_df.dropna(axis="rows", how="any")

        # regrid coordinate: genie lat x normal lon
        ## lat
        if hasattr(self, "lat"):
            src_df.loc[:, self.lat] = src_df.loc[:, self.lat].apply(go.geniebin_lat)
        if hasattr(self, "lon"):
            src_df.loc[:, self.lon] = src_df.loc[:, self.lon].apply(go.normbin_lon)
        if hasattr(self, "depth"):
            src_df.loc[:, self.depth] = src_df.loc[:, self.depth].apply(go.geniebin_depth)

        # aggregated source data (in long format)
        src_df_agg = src_df.groupby(self.index).agg(agg_method)
        return src_df_agg

    def drop_na(self, *args, **kwargs):
        "drop rows with NA values"
        self.data = self.data.dropna(*args, **kwargs)
        return self

    def to_ScatterDataVis(self):
        "convert to ScatterDataVis"
        return ScatterDataVis(self)

    def plot(self, var, kind='scatter',*args, **kwargs):
        "plot the data"
        return self.to_ScatterDataVis().plot(var, kind=kind,*args, **kwargs)

    def compare(self, var1, var2, model_name=None, obs_name=None):
        "compare two ScatterData objects"
        if not model_name:
            model_name = var1            
        if not obs_name:
            obs_name = var2
            
        return DFComparison(self.data, var1, var2, model_name=model_name, obs_name=obs_name)

    def rolling(self, window, *args, **kwargs):
        "apply rolling to the data"
        if self.mutable:
            self.data = self.data.rolling(window, *args, **kwargs)
            return self
        else:
            return self.data.rolling(window, *args, **kwargs)




from cgeniepy.plot import CommunityPalette
import matplotlib.colors


def test_txt_palette():
    cmap = CommunityPalette().get_palette('ODV')
    assert matplotlib.colors.to_hex(cmap.colors[0]) == '#feb483'


def test_spk_palette():
    cmap = CommunityPalette().get_palette('medium_rainbow')
    assert matplotlib.colors.to_hex(cmap.colors[0]) == '#cc00ff'


def test_xml_palette():
    cmap = CommunityPalette().get_palette('Section')
    assert matplotlib.colors.to_hex(cmap.colors[0]) == '#c2ccb8'


def test_alt_init():
    hex_codes = CommunityPalette('medium_rainbow').to_hex()
    assert hex_codes[0] == '#cc00ff'



from cgeniepy.table import ScatterData
import pandas as pd

def create_testdata():
    lat = -48.876
    lon = 123.393
    df = pd.DataFrame({'lat': [lat], 'lon': [lon]})
    return ScatterData(df)

def test_init():
    data = create_testdata()
    ## note the order is put in reverse intentionally
    data.set_index(['lon','lat']) 
    assert data.lat == 'lat'

def test_detectbasin():
    data = create_testdata()
    data.set_index(['lat','lon'])
    basin_value = data.detect_basin()['basin'].values.item()
    assert basin_value =='Indian Ocean'

def test_url():
    url = "https://www.ncei.noaa.gov/pub/data/paleo/icecore/antarctica/epica_domec/edc-co2-2008-bern-noaa.txt"
    
    test_data= ScatterData(url, comment='#', delimiter='\t')
    assert test_data.data['CO2'][0] == 257.8


import cgeniepy

def test_model_getvar():
    model = cgeniepy.sample_model()
    assert model.get_var("ocn_sur_temp").mean().data.values.item() == 18.055744171142578
    
def test_eco_pft():
    model = cgeniepy.sample_model(model_type='EcoModel',gemflag=['ecogem'])
    ## 1st PFT
    data = model.get_pft(1).isel(time=-1).mean().data.values.item()
    assert data  == 0.12259095907211304

def test_eco_multipft():
    model = cgeniepy.sample_model(model_type='EcoModel',gemflag=['ecogem'])
    ## 1st and 2nd PFT
    data = model.get_pft([1,2]).isel(time=-1).sum(dim='variable').mean().data.values.item()
    assert data  == 0.16739805042743683



from cgeniepy.array import GriddedData
import numpy as np
import xarray as xr

def create_testdata():
    lat = np.linspace(-89.5,89.5,180)
    lon = np.linspace(0,359,360)
    np.random.seed(12349)
    data = np.random.rand(lat.size,lon.size)
    xdata = xr.DataArray(data, coords=[('lat',lat),('lon',lon)],
                         attrs={'long_name':'random data', 'units':'uniteless'})   
    return GriddedData(xdata,False, attrs=xdata.attrs)

def test_mean():
    data = create_testdata()
    assert data.mean().data.item() == 0.5021667118489245

def test_sd():
    data = create_testdata()
    assert data.sd().data.item() == 0.28845163932542145

def test_variance():
    data = create_testdata()
    assert data.variance().data.item() == 0.08320434822952301

def test_median():
    data = create_testdata()
    assert data.median().data.item() == 0.5036269709952814


def test_min():
    data = create_testdata()
    assert data.min().data.item() == 1.2270557283589056e-06    

def test_max():
    data = create_testdata()
    assert data.max().data.item() == 0.9999894976042956

def test_search_point():
    data = create_testdata()
    ## nemo point lat/lon
    lat = -48.876
    lon = 123.393
    assert data.search_point((lat,lon), ignore_na=True) == 0.25995689209624817

def test_sel_modern_basin():
    data = create_testdata()
    assert data.sel_modern_basin(50,norm_lon_method='').mean().data.item() == 0.5019781051132972



from typing import Union
import numpy as np
from scipy.spatial import distance
from scipy.stats import pearsonr
import matplotlib.pyplot as plt

from matplotlib.projections import PolarAxes
import mpl_toolkits.axisartist.floating_axes as fa
import mpl_toolkits.axisartist.grid_finder as gf

import warnings

class ArrComparison:
    
    "quantitatively compare similarity metrics between two 2D arrays (model and observation)"

    def __init__(self, model, observation, model_name="Model", obs_name="Observation", label=None):
        """
        
        :param: model: array-like data
        :param: observation: array-like data

        --------
        Example:
        --------
        
        from cgeniepy.skill import ModelSkill

        ## generate 2d data
        import numpy as np
        x = np.random.rand(36, 36)
        y = np.random.rand(36, 36)

        ## calculate skill score
        ms = ModelSkill(x, y)
        ms.mscore()
        """
        self.model = model
        self.data = observation

        self.model_name = model_name
        self.obs_name = obs_name
        self.label = label

        ## check data type, if not float, convert to float
        if self.model.dtype != np.float64:
            self.model = self.model.astype(np.float64)
        if self.data.dtype != np.float64:
            self.data = self.data.astype(np.float64)
        

    def safe_unveil(self, array):
        "get pure array from a numpy masked array object"
        if array.__class__ != np.ma.core.MaskedArray:
            return array
        else:
            return array.filled(np.nan)

    def intersect_index(self, verbose=False):
        """ Return the index where corresponding values are not
        nan in both input arrays. One then can filter the array
        by the output boolean array.
        """

        # If both are not NaN, return it
        array1 = self.safe_unveil(self.model)
        array2 = self.safe_unveil(self.data)

        indx_array = np.logical_and(~np.isnan(array1), ~np.isnan(array2))

        if verbose is True:
            num = indx_array.flatten()[indx_array.flatten() == True].shape[0]
            print("Summary: {} elements simultaneously exist.".format(num))

        return indx_array

    def mscore(self):
        """
        Calculate skill metric M-score. See more in the paper Watterson, I. G. (1996)
        """

        # get common set
        indx = self.intersect_index()
        sub_data1 = self.model[indx]
        sub_data2 = self.data[indx]

        # calculate M-score
        mse = np.square(np.subtract(sub_data1, sub_data2)).mean()
        v1 = sub_data1.var()
        v2 = sub_data2.var()
        g1 = sub_data1.mean()
        g2 = sub_data2.mean()

        mscore = (2 / np.pi) * np.arcsin(1 - mse / (v1 + v2 + np.square(g1 - g2)))

        return mscore

    def pearson_r(self):
        """
        calculate pearson correlation coefficient for 2D array
        """
        # get common set
        indx = self.intersect_index()
        sub_data1 = self.model[indx]
        sub_data2 = self.data[indx]

        # correlation matrix
        corr_mat = np.corrcoef(sub_data1, sub_data2)
        corr = corr_mat[0, 1]

        return corr

    def cos_similarity(self):
        """
        Cosine similarity that ignores the magnitude of data, but focuses on the direction
        """
        # get common set
        indx = self.intersect_index()
        sub_data1 = self.model[indx].ravel()
        sub_data2 = self.data[indx].ravel()

        if sub_data1.mean() != 0 and sub_data2.mean() != 0:
            cos_sim = 1 - distance.cosine(sub_data1.flatten(), sub_data2.flatten())
        else:
            cos_sim = np.nan

        return cos_sim

    def rmse(self):
        """
        Root Mean Sqaure Error (rmse, or rmsd)
        """
        data1 = self.model
        data2 = self.data

        error_2d = data1 - data2
        error_1d = error_2d.ravel()[~np.isnan(error_2d.ravel())]
        rmse = np.sqrt(np.square(error_1d).mean())

        return rmse

    def nrmse(self):
        "Normalised RMSE facilitating the comparison between datasets or models with different scales"

        norm_rmse = self.rmse()/np.nanmean(self.model)
        return norm_rmse

    def crmse(self):
        """
        centred Root Mean Sqaure Error (rmse, or rmsd). See Talor, K. E. (2001) JGR
        """
        # select data
        indx = self.intersect_index()
        sub_data1 = self.model[indx].ravel()
        sub_data2 = self.data[indx].ravel()

        # calculate std
        sigma1 = np.std(sub_data1)
        sigma2 = np.std(sub_data2)

        # pearson correlation
        corr_mat = np.corrcoef(sub_data1, sub_data2)
        corr = corr_mat[0, 1]

        # central rmse
        crmse = sigma1**2 + sigma2**2 - 2 * sigma1 * sigma2 * corr

        return crmse
    
    def lm_pvalue(self):
        """
        get p-value from linear regression
        """
        
        indx = self.intersect_index()
        sub_data1 = self.model[indx].ravel()
        sub_data2 = self.data[indx].ravel()        
        
        r, p = pearsonr(sub_data1, sub_data2)
        
        return p

    def plot(self, ax=None, diagonal=True,title="Model vs Observation", savefig_name=None, metric=['rmse', 'r2', 'm', 'p'], *args, **kwargs):
        ## a x-y plot
        if not ax:
            fig, ax = plt.subplots()
        plt.rcParams['font.family'] = 'sans-serif'
        ax.grid(True, linestyle='--', alpha=0.5)
        ax.scatter(self.model, self.data, zorder=2,*args, **kwargs)
        ax.minorticks_on()
        ax.tick_params(axis='both', which='both', direction='in', top=True, right=True)
        ax.set_xlabel(self.model_name)
        ax.set_ylabel(self.obs_name)
        ## add 1:1 line
        if diagonal:
            ## add real 1:1 line
            ax.plot([self.model.min(), self.model.max()], [self.model.min(), self.model.max()], color='k', ls='--')
        
        ## add metrics
        if 'rmse' in metric:
            ax.text(0.05, 0.8, "RMSE: {:.3f}".format(self.rmse()), transform = ax.transAxes)
            
        if 'r2' in metric:                 
             ax.text(0.05, 0.85, "Pearson R: {:.3f}".format(self.pearson_r()),transform = ax.transAxes)
                
        if 'm' in metric:
            ax.text(0.05, 0.75, "M-score: {:.3f}".format(self.mscore()),transform = ax.transAxes)
            
        if 'p' in metric:
            ax.text(0.05, 0.9, "p-value: {:.4f}".format(self.lm_pvalue()),transform = ax.transAxes)
            
            
        
                

        ax.set_title(title, loc='left', fontweight='bold', fontsize=12)

        if savefig_name:
            fig.savefig(savefig_name)

    
class DFComparison(ArrComparison):

    """
    Quantitatively compare similarity metrics between two columns in a dataframe-like objects
    """

    def __init__(self, df, model_col, observation_col, *args, **kwargs):
        self.model = df[model_col].to_numpy()
        self.data = df[observation_col].to_numpy()
        super().__init__(self.model, self.data, *args, **kwargs)

 
class TaylorDiagram(object):
    """
    Taylor diagram.

    A visualisation of model-data comparison considering (1) RMSE; (2) Correlation; (3) Standard Deviation.
    
    In practical it is a polar axis with point with `std` as radiance and `corr` as theta.
    The cRMSE then measures the distance from reference point, which is at (ref_std, 0)

    see https://en.wikipedia.org/wiki/Taylor_diagram#/media/File:Taylor_diagram_fig2.png
    and https://bookdown.org/david_carslaw/openair/sections/model-evaluation/taylor-diagram.html (Figure 20.2)

    This class is modified from Yannick Copin, https://gist.github.com/ycopin/3342888
    """

    def __init__(self, ac: Union[ArrComparison,list]):
        if isinstance(ac, list):
            self.mult_comp = True
        else:
            self.mult_comp = False
            
        self.ac = ac
        self._extract_data()

    def _extract_data(self):

        if not self.mult_comp:        
            self.corr = self.ac.pearson_r()
            self.model_std = np.std(self.ac.model)
            self.obs_std = np.std(self.ac.data)
            self.label = self.ac.label
        else:
            corr, model_std, obs_std, label = [],[],[],[]
            for ac_i in self.ac:
                corr.append(ac_i.pearson_r())
                model_std.append(np.std(ac_i.model))
                obs_std.append(np.std(ac_i.data))
                label.append(ac_i.label)                

            self.corr = corr
            self.model_std = model_std
            self.obs_std = obs_std
            self.label = label
                

    def setup_ax(self, fig=None, positive_only=True, crmse_contour=False):

        tr = PolarAxes.PolarTransform()

        ## 1.creating extrems for gridhelper
          ## ymax = 1/2 pi if
        if positive_only:
            ymax= 1/2 * np.pi
        else:
            ymax = np.pi

        x0 = 0
        if self.mult_comp:
            self.ref_std = 1
        else:
            self.ref_std = self.obs_std

        x1 = self.ref_std * 1.6
        extremes=(0, ymax, x0, x1)

        # 2. grid locator     
        rlocs = np.array([0, 0.2, 0.4, 0.6, 0.8, 0.9, 0.95, 0.99, 1])
        # a concatenate with ticks' reverse
        rlocs = np.concatenate((-rlocs[:0:-1], rlocs))
        ## arc-cos -> convert to polar angles
        tlocs = np.arccos(rlocs)
        grid_locator1 = gf.FixedLocator(tlocs)    # Positions

        # 3. tick formatters for each axis
        tf1 = gf.DictFormatter(dict(zip(tlocs, map(str, rlocs))))

        ## gridhelper to create the special axis
        gridhelper = fa.GridHelperCurveLinear(
                tr, extremes=extremes,
                grid_locator1=grid_locator1,
                tick_formatter1=tf1)

        ## create figure
        if not fig:
            self.fig = plt.figure()
        else:
            self.fig = fig
            
        ax = self.fig.add_axes(111, axes_class=fa.FloatingAxes, grid_helper=gridhelper)

        # Adjust axes
        plt.rcParams['font.family'] = 'sans-serif'
        ax.axis["top"].set_axis_direction("bottom")   # "Angle axis"
        ax.axis["top"].toggle(ticklabels=True, label=True)
        ax.axis["top"].major_ticklabels.set_axis_direction("top")
        ax.axis["top"].label.set_axis_direction("top")
        ax.axis["top"].label.set_text("Correlation")

        ax.axis["left"].set_axis_direction("bottom")  # "X axis"
        ax.axis["left"].label.set_text("Standard deviation")

        ax.axis["right"].set_axis_direction("top")    # "Y-axis"
        ax.axis["right"].toggle(ticklabels=True)

        ax.axis["bottom"].toggle(ticklabels=False, label=False)

        ## add grid lines
        ax.grid(linestyle='dashed', axis='x')


        self._ax = ax # Graphical axes
        self.ax = ax.get_aux_axes(tr) # Polar coordinates

        if crmse_contour:
            rs, ts = np.meshgrid(np.linspace(x0, x1), np.linspace(0, ymax))
            RMSE=np.sqrt(np.power(self.ref_std, 2) + np.power(rs, 2) - (2.0 * self.ref_std * rs  *np.cos(ts)))
            contours = self.ax.contour(ts, rs, RMSE, 5, linestyles='dashed', colors='#d4af37', linewidths=1.5,
                                       zorder=1)
            ## add label
            self.ax.clabel(contours, inline=True, fontsize=10)

            ## add reference std line
            theta_ref = np.linspace(0, 2*np.pi, 100)
            radiance_ref = theta_ref*0 + self.ref_std
            self.ax.plot(theta_ref, radiance_ref, color='lightblue', linestyle='dashed',
                         linewidth=1.5)

            ## add reference point (because crmse is the distance to reference point)
            self.add_point(1, self.ref_std, marker='*',label='reference', edgecolor='k')            


    def add_point(self, correlation, std, *args, **kwargs):
        ## x->theta, y->radiance
        self.ax.scatter(np.arccos(correlation), std,*args, **kwargs)        

    def plot(self):
        if not self.mult_comp:
            ## add model point
            self.add_point(self.corr, self.model_std, edgecolor='k', label=self.label)
        else:
            
            for i in range(len(self.ac)):
                ## normalised std
                self.add_point(self.corr[i], self.model_std[i]/self.obs_std[i], label=self.label[i],
                               edgecolor='k')
        self.ax.legend()

    def savefig(self, *args, **kwargs):
        self.fig.savefig(*args, **kwargs)

    def add_line(self,theta1, radius1,theta2, radius2, *args, **kwargs):
        """
        Add a line in the polar axis

        If you need to convert the correlation into theta/radian, use np.arccos(correlation)
        """
        line_2d = plt.Line2D([theta1,theta2], [radius1,radius2])
        self.ax.add_line(line_2d, *args, **kwargs)



# from pint import UnitRegistry
# import pathlib


# ureg = UnitRegistry()
# Q_ = ureg.Quantity
# file_path = pathlib.Path(__file__).parent.parent / "data/context.txt"
# ureg.load_definitions(file_path)


from importlib.resources import files
from cgeniepy.model import GenieModel
from cgeniepy.ecology import EcoModel

def sample_model(model_type='GenieModel', *args, **kwargs):
    file_path=str(files('data').joinpath('sample_model'))
    
    if model_type == 'GenieModel':
        model = GenieModel(file_path,*args, **kwargs)
    elif model_type == 'EcoModel':
        model = EcoModel(file_path,*args, **kwargs)
    else:
        raise ValueError('model_type must be either GenieModel or EcoModel')    

    return model
    



from cgeniepy.grid import GridOperation as go
import numpy as np
def test_lon_n2g():
    assert go().lon_n2g(100, -270) == -260


def test_lon_g2n():
    assert go().lon_g2n(-260) == 100


def test_lon_e2n():
    assert go().lon_e2n(350) == -10

def test_lon_n2e():
    assert go().lon_n2e(-10) == 350


def test_geodistance_2d():
    pnt1=(0, 0,0)
    pnt2= np.array([[0, 10, 10]])

    assert go().geo_dis2d(pnt1, pnt2).item() == 1111.9492664455872

def test_checkdimension():
    input = ['lat','lon','time','dpeth'] ## intentional typo
    has_lat, has_lon, has_depth, has_time = go().check_dimension(input)
    assert has_depth == False


def test_dimorder():
    input = ['lat','lon','time','depth']
    depth_order = go().dim_order(input)[1]
    assert input[depth_order] == 'depth'


import numpy as np
import xarray as xr
from scipy.stats import sem
import regionmask

from cgeniepy.grid import Interpolator, GridOperation
from cgeniepy.plot import GriddedDataVis
from cgeniepy.chem import Chemistry
import cgeniepy.table as ct
from functools import cache

import warnings


class GriddedData:
    """
    GriddedData is a class to store and compute GENIE netcdf data.

    It stores data in xarray.DataArray format, and provides optimalised methods for GENIE model output to compute statistics.    
    """
    
    def __init__(self, array=np.nan, mutable=False, attrs={}):
        """
        Initialise an instance of GriddedData

        
        :param array: a xarray.DataArray object
        :param mutable: whether the data is mutable in place, default is False
        :param attrs: a dictionary of attributes, default is empty

        -----------
        Example
        -----------
        data = xr.DataArray(np.random.randn(2, 3), dims=("x", "y"), coords={"x": [10, 20]}, attrs={"units": "unitless""})
        dg = GriddedData(data, mutable=False)
        """
        
        # set data
        self.data = array
        self.mutable = mutable
        self.attrs = attrs

        ## formatting the unit
        if 'units' in self.attrs:
            self.attrs['units'] = Chemistry().format_unit(self.attrs['units'])

    def __repr__(self):
        return f"GriddedData\ndata={self.data}\nmutable:\n    {self.mutable}"

    def __getitem__(self, item):
        "make GriddedData subscriptable like xarray.DataArray"
        return self.data[item]

    def __array__(self):
        "return the numpy array of the data"
        return self.data.values

    def interpolate(self, *args, **kwargs):
        """Interpolate the GriddedData to a finer grid, mostly useful for GENIE plotting

        :returns: a GriddedData object with interpolated data
        
        ------------
        Example
        ------------
        >>> model = GenieModel(path)
        >>> var = model.get_var(target_var)
        >>> var.interpolate(method='r-linear') ## regular linear interpolation        
        """
        
        coords = tuple([self.data[dim].values for dim in self.data.dims])
        values = self.data.values
        dims = self.data.dims
        output = Interpolator(dims, coords,values, *args, **kwargs).to_xarray()
        
        if self.mutable:
            self.data = output
            self.attrs = self.attrs            
            return self            
        else:
            return GriddedData(output, mutable=False, attrs=self.attrs)
        

    def sel(self, *args, **kwargs):
        """select grid points based on the given coordinates
        a wrapper to xarray `sel` method

        :returns: a GriddedData object with selected data

        -----------
        Examples:
        -----------
        >>> model = GenieModel(path)
        >>> var = model.get_var(target_var)
        >>> var.search_grid(lon=XX, lat=XX, zt=XX, method='nearest')
        """

        if self.mutable:
            self.data = self.data.sel(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.sel(*args, **kwargs))


    def isel(self, *args, **kwargs):
        """select grid points based on the given index
        a wrapper to xarray `isel` method

        -----------
        Examples:
        -----------
        >>> model = GenieModel(path)
        >>> var = model.get_var(target_var)
        >>> var.isel(lon=XX, lat=XX, zt=XX)
        """
        if self.mutable:
            self.data = self.data.isel(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.isel(*args, **kwargs), mutable=False, attrs=self.attrs)
    
    def normalise_longitude(self, method='g2n', *args, **kwargs):
        """Normalise GENIE's longitude (eastern degree) to normal longitude (-180, 180)

        :param method: normalise method, default is 'g2n' (GENIE to normal). A full list of methods are:
            - 'g2n': GENIE to normal
            - 'n2g': normal to GENIE
            - 'e2n': eastern to normal
            - 'n2e': normal to eastern

        -----------
        Example
        -----------

        >>> Model = GenieModel("a path")
        >>> Model.get_var('abc').normalise_longitude()
        """
        gp = GridOperation()

        match method:
            case 'g2n':
                output = gp.xr_g2n(self.data, *args, **kwargs)
            case 'n2g':
                output = gp.xr_n2g(self.data, *args, **kwargs)
            case 'e2n':
                output = gp.xr_e2n(self.data, *args, **kwargs)
            case 'n2e':
                output = gp.xr_n2e(self.data, *args, **kwargs)
            case _:
                ## No change
                warnings.warn("No change applied because of invalid method")
                output = self.data

        if self.mutable:
            self.data = output
            return self        
        else:
             return GriddedData(output, mutable=False, attrs=self.attrs)
                    

    def __add__(self, other):
        """Allow GriddedData to be added by a number or another GriddedData
        """
        sum_array = GriddedData()
        if hasattr(other, "array"):
            sum_array.data = self.data + other.array
            ## check the attributes
            np.testing.assert_equal(self.attrs, other.attrs)            
            sum_array.attrs = self.attrs
        else:
            ## a scalar or xarray.DataArray
            sum_array.data = self.data + other
            sum_array.attrs = self.attrs
        return sum_array

    def __radd__(self, other):
        """
        Allow a number or another GriddedData to be added to this GriddedData
        """
        sum_array = GriddedData()
        if hasattr(other, "data"):
            sum_array.data = other.data + self.data
            ## check all the attributes
            np.testing.assert_equal(self.attrs, other.attrs)
            sum_array.attrs = self.attrs
        else:
            sum_array.data = other + self.data
            sum_array.attrs = self.attrs
        return sum_array

    def __sub__(self, other):
        """
        Allow GriddedData to be subtracted by a number or another GriddedData
        """
        diff = GriddedData()
        if hasattr(other, "data"):
            
            diff.data = self.data - other.data
            ## check all the attributes
            np.testing.assert_equal(self.attrs, other.attrs)
            diff.attrs = self.attrs
        else:
            ## a scalar
            diff.data = self.data - other
        return diff

    def __rsub__(self, other):
        """
        Allow a number or another GriddedData to be subtracted from this GriddedData
        """
        diff = GriddedData()
        if hasattr(other, "data"):
            diff.data = other.data - self.data
            # check all the attributes
            np.testing.assert_equal(self.attrs, other.attrs)
            diff.attrs = self.attrs
        else:
            # a scalar
            diff.data = other - self.data

        return diff

    def __truediv__(self, other):
        """
        Allow GriddedData to be divided by a number or another GriddedData

        NA/NA -> NA
        """
        quotient_array = np.zeros_like(self.data)
        
        if isinstance(other, GriddedData):
            quotient_array = np.divide(self.data, other.data)
        else:
            # Handle division by a scalar
            quotient_array = np.divide(self.data, other)

        quotient = GriddedData()
        quotient.data = xr.DataArray(quotient_array)
        return quotient

    def __rtruediv__(self, other):
        """
        Allow a number or another GriddedData to be divided by this GriddedData
        """
        quotient_array = np.zeros_like(self.data)
        if isinstance(other, GriddedData):
            quotient_array = np.divide(other.data, self.data)
        else:
            # Handle division by a scalar
            quotient_array = np.divide(other, self.data)

        quotient = GriddedData()
        quotient.data = xr.DataArray(quotient_array)
        return quotient

    def __mul__(self, other):
        """
        Allow GriddedData to be multiplied by a number or another GriddedData
        """
        product = GriddedData()
        if hasattr(other, "data"):
            product.data = self.data * other.data
        else:
            try:
                product.data = self.data * other
            except ValueError:
                print("Sorry, only number and GriddedData are accepted")

        return product

    def __rmul__(self, other):
        """
        Allow a number or another GriddedData to be multiplied by this GriddedData
        """
        product = GriddedData()
        if hasattr(other, "data"):
            product.data = other.data * self.data
        else:
            try:
                product.data = other * self.data
            except ValueError:
                print("Sorry, only number and GriddedData are accepted")

        return product

    def __pow__(self, other):
        """
        Allow GriddedData to be raised to a power
        """

        if self.mutable:
            self.data = self.data ** other
            return self
        else:
            return GriddedData(self.data ** other, mutable=False, attrs=self.attrs)


    ## allow comparison
    def __lt__(self, other):
        return self.data < other

    def __le__(self, other):
        return self.data <= other

    def __gt__(self, other):
        return self.data > other
            
    def __ge__(self, other):
        return self.data >= other

    def __eq__(self, other):
        return self.data == other
            
    def __ne__(self, other):
        return not self.__eq__(other)


    def max(self, *args, **kwargs):
        """compute the maximum value of the array
        """
        
        self.data = self.data.max(*args, **kwargs)
        return self


    def min(self, *args, **kwargs):
        """compute the minimal value of the array
        """        
        if self.mutable:
            self.data = self.data.min(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.min(*args, **kwargs), mutable=False, attrs=self.attrs)


    def sum(self, *args, **kwargs):
        """ compute the sum of the array
        """
        
        if self.mutable:
            self.data = self.data.sum(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.sum(*args, **kwargs), mutable=False, attrs=self.attrs)


    def mean(self, *args, **kwargs):
        """compute the mean of the array.
        Note this is not weighted mean, for weighted mean, use `weighted_mean` method
        """
        if self.mutable:
            self.data = self.data.mean(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.mean(*args, **kwargs), mutable=False, attrs=self.attrs)
    

    def median(self, *args, **kwargs):
        """compute the median of the array
        """
        
        if self.mutable:
            self.data = self.data.median(*args, **kwargs)
            return self
        else:
            return GriddedData(self.data.median(*args, **kwargs), mutable=False, attrs=self.attrs)

    def sd(self, *args, **kwargs):        
        "compute the standard deviation of the mean"
        if self.mutable:
            self.data = np.std(self.data, *args, **kwargs)
            return self
        else:
            return np.std(self.data, *args, **kwargs)


    def variance(self, *args, **kwargs):
        "compute the variance of the array"
        if self.mutable:
            self.data = np.var(self.data, *args, **kwargs)
            return self
        else:
            return np.var(self.data, *args, **kwargs)

    def se(self, *args, **kwargs):
        "compute the standard error of the mean"
        if self.mutable:
            self.data = sem(self.data, nan_policy="omit", axis=None, *args, **kwargs)
            return self
        else:
            return sem(self.data, nan_policy="omit", axis=None, *args, **kwargs)

    def weighted_mean(self, weights, *args, **kwargs):
        """
        compute the weighted average of the array (e.g., ocean volume)
        """
        array_ma = self.data.to_masked_array()
        return np.ma.average(array_ma, weights=weights, *args, **kwargs)        
    
    def sel_modern_basin(self, basin, norm_lon_method='g2n'):
        """
        select modern basin from regionmask.defined_regions.ar6.ocean

        The 58 defines marine regions from the sixth IPCC assessment report (AR6), Iturbide et al., (2020) ESSD
        
        https://regionmask.readthedocs.io/en/stable/_images/plotting_ar6_all.png

        :param basin: the basin index (int) or basin abbrev name (str), or a list of basin index or basin name
        :param norm_lon_method: normalise longitude method, default is 'g2n'
        
        --------------------------------
        47: North Pacific
        48: Equatorial Pacific
        49: South Pacific
        
        50: North Atlantic Ocean
        51: Equatorial Atlantic Ocean
        52: Southern Atlantic Ocean

        53: North Indian Ocean
        55: Equatorial Indian Ocean
        56: South Indian Ocean

        57: Southern Ocean
        46: Arctic Ocean
        --------------------------------

        
        Example:
        ---------        
        >>> gd.sel_modern_basin(47) ## North Pacific
        """
        ocean = regionmask.defined_regions.ar6.ocean

        target_ocean_index = ocean.map_keys(basin)
        mask = ocean.mask(self.normalise_longitude(method=norm_lon_method)) ## an masked array target regions are the index, others are NA

        if isinstance(target_ocean_index, list):
            cond_lst = []
            for i in target_ocean_index:
                cond_lst.append(mask == i)
            cond = np.logical_or.reduce(cond_lst)
        else:
            cond = mask == target_ocean_index


        if self.mutable:
            self.data = self.normalise_longitude(method=norm_lon_method).data.where(cond)                    
            return self
        else:
            return GriddedData(self.normalise_longitude(method=norm_lon_method).data.where(cond), mutable=False, attrs=self.attrs)


    def mask_basin(self, base, basin, subbasin):

        """use pre-defined grid mask to select_basin, mostly used for cGENIE model

        :param base: the base configuration name, e.g., worjh2, worlg4
        :param basin: the basin name, e.g., Atlantic, Pacific, Indian
        :param subbasin: N/S/ALL, ALL means Southern Ocean section included

        Example
        ---------
        >>> gd.mask_basin('worjh2', 'Atlantic'').mean()        
        """
        gp = GridOperation()
        data = self.data
        attr = self.attrs
        mask = gp.GENIE_grid_mask(base=base, basin=basin, subbasin=subbasin, invert=True)

        if self.data.ndim > 2:
            mask = np.broadcast_to(mask, (16, 36, 36))

        mask_data = np.ma.array(data, mask=mask)
        mask_data = np.ma.masked_invalid(mask_data)

        output = xr.DataArray(mask_data, dims=data.dims, coords=data.coords)


        if self.mutable:
            self.data = output
            self.attrs = attr        
            return self
        else:
            return GriddedData(output, mutable=False, attrs=attr)


    @cache
    def ocn_only_data(self, index=False):
        """
        remove the NA grid (i.e., land definition)
        :param index: return GeoIndex if True or real value if false
        """

        stacked_data = self.data.stack(x=self.data.dims)            
        stacked_ocn_mask =~np.isnan(stacked_data)
        ocn_only_data = stacked_data[stacked_ocn_mask]
        
        if index:
            return np.stack(ocn_only_data.x.values)
        else:
            return ocn_only_data
        

    def search_point(self, point, ignore_na=False, to_genielon=False, **kwargs):
        """
        search the nearest grid point to the given coordinates
        
        :param point: a list/tuple of coordinate values in the same order to the data dimension (self.data.dims)
        :param ignore_na: whether only check ocean data (which ignore the NA grids)
        :param to_genielon: whether convert the input longitude to genie longitude, input point must be list if True

        ----------------------
        Example
        ----------------------
        >>> lat, lon, depth = 10, 20, 30
        >>> data.search_point((lat, lon, depth))
        """
        ## ignore the first dimension (which is time)        
        ndim = self.data.ndim
        
        if len(point) != ndim:
            raise ValueError("Input point has incompatiable coordinate")
        
        if ignore_na:
            index_pool = self.ocn_only_data(index=True)
            match ndim:
                case 3:
                    ## point: (depth, lat, lon); Index: (depth, lat, lon)
                    if to_genielon: point[2] = GridOperation().lon_g2n(point[2])
                    distances = GridOperation().geo_dis3d(point, index_pool[:, 0:4])
                    idx_min = np.argmin(distances)
                case 2:
                    ## point: (lat, lon); Index: (lat, lon)
                    if to_genielon: point[1] = GridOperation().lon_g2n(point[1])
                    distances = GridOperation().geo_dis2d(point, index_pool[:, 0:3])
                    idx_min = np.argmin(distances)
            
            nearest_value = self.ocn_only_data(index=False).values[idx_min]
            return nearest_value
        else:
            ## construct a dictionary based on self.data.dims and input point, and method ("nearest")
            kwargs = dict(zip(self.data.dims, point), method='nearest')
            return self.data.sel(**kwargs).values.item()

    def to_GriddedDataVis(self):
        """
        convert GriddedData to GriddedDataVis
        """
        return GriddedDataVis(self)

    def plot(self, *args, **kwargs):
        """
        plot the data using GriddedDataVis
        """
        return self.to_GriddedDataVis().plot(*args, **kwargs)

    def to_dataframe(self):
        """
        convert GriddedData to pandas.DataFrame
        """
        return self.data.to_dataframe()

    def to_ScatterData(self):
        """
        convert GriddedData to ScatterData
        """
        return ct.ScatterData(self.to_dataframe())
